function FPP = FastProlatePseudoinverse(N,W,cutoff,epsilon)
%% FastProlatePseudoinverse.m 
% Function implementing the fast truncated pseudoinverse of B_{N,W} as 
% described in "The Fast Slepian Transform". The rank K \approx 2NW
% truncated pseudoinverse of the prolate matrix can be approximated as the
% the prolate matrix plus a low rank correction matrix.
% 
% Inputs:
% -N: Length of DPSSs
% -W: Half Bandwidth for DPSSs
% -cutoff: rank K = cutoff
%          OR rank K = #{ell : lambda_{N,W}^{(ell)} >= cutoff} DPSSs.
% -epsilon: Relative error in approximation
%
% Outputs:
% -FPP an object containing a Fast Prolate Pseudoinverse function
%
% Given an Nx1 vector x, performing
% z = FPP.FPpinv(x);
% results in z being an Nx1 vector such that ||z-B_{N,W}^{\dagger}x||_2 <= epsilon*||x||_2
% If x is an NxM array, then FPP.FPpinv() operates columnwise.
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

%% Fast Slepian Projection Function
FPP.FPpinv = @FPpinv;

%% Compute DPSSs and store transition region DPSSs
[FPP.S,lambda,lowerindex,upperindex] = transitionDPSS(N,W,[epsilon;(2+epsilon-sqrt(4+epsilon^2))/2],cutoff);
if(cutoff >= 1)
    K = cutoff;
    FPP.eig_weights = [1./lambda(1:(K-lowerindex+1));zeros(upperindex-K,1)]-lambda;
elseif(cutoff > eps)
    FPP.eig_weights = (lambda > cutoff)./lambda-lambda;
    K = nnz(lambda > cutoff)+lowerindex-1;
end

%% Precompuute length 2^q FFT of sinc function needed to apply B_{N,W} to a vector (where 2^q >= 2*N-1).
FPP.fftlength = 2^ceil(log2(2*N-1));
vecSinc = 2*W*sinc(2*W*(0:(N-1))');
FPP.sincFFT = fft([vecSinc;...
                   zeros(FPP.fftlength-2*N+1,1);...
                   vecSinc(end:-1:2)]);

%% Store signal length, bandwidth, tolerance, and number of DPSSs
FPP.N = N; % Signal Length
FPP.W = W; % Half-Bandwidth
FPP.eps = epsilon; % Tolerance
FPP.cutoff = cutoff;
FPP.K = K; % Number of DPSSs

%% Fast Prolate Pseudoinverse
function Bpinvx = FPpinv(x)
    if(FPP.N ~= size(x,1))
        disp('ERROR: Input has incorrect length')
    else
        % Toeplitz matrix multiply (Use FFT of length 2^q >= 2*N-1)
        Bx = ifft(bsxfun(@times,FPP.sincFFT,fft(x,FPP.fftlength,1)));
        
        % Low rank update
        SDSx = (FPP.S)*bsxfun(@times,FPP.eig_weights,(FPP.S)'*x);
        
        % Add Toeplitz*vector to low rank update
        Bpinvx = Bx(1:FPP.N,:)+SDSx;
    end
end

end

